# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

import re  # noqa: F401

# python 2 and python 3 compatibility library
import six

from lightly.openapi_generated.swagger_client.api_client import ApiClient


class EmbeddingsApi(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def delete_embedding_by_id(self, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """delete_embedding_by_id  # noqa: E501

        Deletes a embedding entry by id.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.delete_embedding_by_id(dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.delete_embedding_by_id_with_http_info(dataset_id, embedding_id, **kwargs)  # noqa: E501
        else:
            (data) = self.delete_embedding_by_id_with_http_info(dataset_id, embedding_id, **kwargs)  # noqa: E501
            return data

    def delete_embedding_by_id_with_http_info(self, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """delete_embedding_by_id  # noqa: E501

        Deletes a embedding entry by id.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.delete_embedding_by_id_with_http_info(dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['dataset_id', 'embedding_id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method delete_embedding_by_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `delete_embedding_by_id`")  # noqa: E501
        # verify the required parameter 'embedding_id' is set
        if self.api_client.client_side_validation and ('embedding_id' not in params or
                                                       params['embedding_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `embedding_id` when calling `delete_embedding_by_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501
        if 'embedding_id' in params:
            path_params['embeddingId'] = params['embedding_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/v1/datasets/{datasetId}/embeddings/{embeddingId}', 'DELETE',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type=None,  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def get_embeddings_by_dataset_id(self, dataset_id, **kwargs):  # noqa: E501
        """get_embeddings_by_dataset_id  # noqa: E501

        Get all annotations of a dataset  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_by_dataset_id(dataset_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :return: list[DatasetEmbeddingData]
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_embeddings_by_dataset_id_with_http_info(dataset_id, **kwargs)  # noqa: E501
        else:
            (data) = self.get_embeddings_by_dataset_id_with_http_info(dataset_id, **kwargs)  # noqa: E501
            return data

    def get_embeddings_by_dataset_id_with_http_info(self, dataset_id, **kwargs):  # noqa: E501
        """get_embeddings_by_dataset_id  # noqa: E501

        Get all annotations of a dataset  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_by_dataset_id_with_http_info(dataset_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :return: list[DatasetEmbeddingData]
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['dataset_id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_embeddings_by_dataset_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `get_embeddings_by_dataset_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/v1/datasets/{datasetId}/embeddings', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='list[DatasetEmbeddingData]',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def get_embeddings_by_sample_id(self, dataset_id, sample_id, **kwargs):  # noqa: E501
        """get_embeddings_by_sample_id  # noqa: E501

        Get all embeddings of a datasets sample  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_by_sample_id(dataset_id, sample_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID sample_id: ObjectId of the sample (required)
        :param str mode: if we want everything (full) or just the summaries
        :return: list[EmbeddingData]
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_embeddings_by_sample_id_with_http_info(dataset_id, sample_id, **kwargs)  # noqa: E501
        else:
            (data) = self.get_embeddings_by_sample_id_with_http_info(dataset_id, sample_id, **kwargs)  # noqa: E501
            return data

    def get_embeddings_by_sample_id_with_http_info(self, dataset_id, sample_id, **kwargs):  # noqa: E501
        """get_embeddings_by_sample_id  # noqa: E501

        Get all embeddings of a datasets sample  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_by_sample_id_with_http_info(dataset_id, sample_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID sample_id: ObjectId of the sample (required)
        :param str mode: if we want everything (full) or just the summaries
        :return: list[EmbeddingData]
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['dataset_id', 'sample_id', 'mode']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_embeddings_by_sample_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `get_embeddings_by_sample_id`")  # noqa: E501
        # verify the required parameter 'sample_id' is set
        if self.api_client.client_side_validation and ('sample_id' not in params or
                                                       params['sample_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `sample_id` when calling `get_embeddings_by_sample_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501
        if 'sample_id' in params:
            path_params['sampleId'] = params['sample_id']  # noqa: E501

        query_params = []
        if 'mode' in params:
            query_params.append(('mode', params['mode']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/users/datasets/{datasetId}/samples/{sampleId}/embeddings', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='list[EmbeddingData]',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def get_embeddings_csv_read_url_by_id(self, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """get_embeddings_csv_read_url_by_id  # noqa: E501

        Get the url of a specific embeddings CSV  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_csv_read_url_by_id(dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: str
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_embeddings_csv_read_url_by_id_with_http_info(dataset_id, embedding_id, **kwargs)  # noqa: E501
        else:
            (data) = self.get_embeddings_csv_read_url_by_id_with_http_info(dataset_id, embedding_id, **kwargs)  # noqa: E501
            return data

    def get_embeddings_csv_read_url_by_id_with_http_info(self, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """get_embeddings_csv_read_url_by_id  # noqa: E501

        Get the url of a specific embeddings CSV  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_csv_read_url_by_id_with_http_info(dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: str
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['dataset_id', 'embedding_id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_embeddings_csv_read_url_by_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `get_embeddings_csv_read_url_by_id`")  # noqa: E501
        # verify the required parameter 'embedding_id' is set
        if self.api_client.client_side_validation and ('embedding_id' not in params or
                                                       params['embedding_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `embedding_id` when calling `get_embeddings_csv_read_url_by_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501
        if 'embedding_id' in params:
            path_params['embeddingId'] = params['embedding_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/v1/datasets/{datasetId}/embeddings/{embeddingId}/readCSVUrl', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='str',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def get_embeddings_csv_write_url_by_id(self, dataset_id, **kwargs):  # noqa: E501
        """get_embeddings_csv_write_url_by_id  # noqa: E501

        Get the signed url to upload an CSVembedding to for a specific dataset  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_csv_write_url_by_id(dataset_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param str name: the sampling requests name to create a signed url for
        :return: WriteCSVUrlData
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.get_embeddings_csv_write_url_by_id_with_http_info(dataset_id, **kwargs)  # noqa: E501
        else:
            (data) = self.get_embeddings_csv_write_url_by_id_with_http_info(dataset_id, **kwargs)  # noqa: E501
            return data

    def get_embeddings_csv_write_url_by_id_with_http_info(self, dataset_id, **kwargs):  # noqa: E501
        """get_embeddings_csv_write_url_by_id  # noqa: E501

        Get the signed url to upload an CSVembedding to for a specific dataset  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.get_embeddings_csv_write_url_by_id_with_http_info(dataset_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param str name: the sampling requests name to create a signed url for
        :return: WriteCSVUrlData
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['dataset_id', 'name']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_embeddings_csv_write_url_by_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `get_embeddings_csv_write_url_by_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501

        query_params = []
        if 'name' in params:
            query_params.append(('name', params['name']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/v1/datasets/{datasetId}/embeddings/writeCSVUrl', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='WriteCSVUrlData',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def set_embeddings_is_processed_flag_by_id(self, body, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """set_embeddings_is_processed_flag_by_id  # noqa: E501

        Sets the isProcessed flag of the specified embedding  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.set_embeddings_is_processed_flag_by_id(body, dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param EmbeddingIdIsProcessedBody body: (required)
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.set_embeddings_is_processed_flag_by_id_with_http_info(body, dataset_id, embedding_id, **kwargs)  # noqa: E501
        else:
            (data) = self.set_embeddings_is_processed_flag_by_id_with_http_info(body, dataset_id, embedding_id, **kwargs)  # noqa: E501
            return data

    def set_embeddings_is_processed_flag_by_id_with_http_info(self, body, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """set_embeddings_is_processed_flag_by_id  # noqa: E501

        Sets the isProcessed flag of the specified embedding  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.set_embeddings_is_processed_flag_by_id_with_http_info(body, dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param EmbeddingIdIsProcessedBody body: (required)
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['body', 'dataset_id', 'embedding_id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method set_embeddings_is_processed_flag_by_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `set_embeddings_is_processed_flag_by_id`")  # noqa: E501
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `set_embeddings_is_processed_flag_by_id`")  # noqa: E501
        # verify the required parameter 'embedding_id' is set
        if self.api_client.client_side_validation and ('embedding_id' not in params or
                                                       params['embedding_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `embedding_id` when calling `set_embeddings_is_processed_flag_by_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501
        if 'embedding_id' in params:
            path_params['embeddingId'] = params['embedding_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/v1/datasets/{datasetId}/embeddings/{embeddingId}/isProcessed', 'PUT',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type=None,  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def trigger2d_embeddings_job(self, body, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """trigger2d_embeddings_job  # noqa: E501

        Trigger job to get 2d embeddings from embeddings  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.trigger2d_embeddings_job(body, dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param Trigger2dEmbeddingJobRequest body: (required)
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.trigger2d_embeddings_job_with_http_info(body, dataset_id, embedding_id, **kwargs)  # noqa: E501
        else:
            (data) = self.trigger2d_embeddings_job_with_http_info(body, dataset_id, embedding_id, **kwargs)  # noqa: E501
            return data

    def trigger2d_embeddings_job_with_http_info(self, body, dataset_id, embedding_id, **kwargs):  # noqa: E501
        """trigger2d_embeddings_job  # noqa: E501

        Trigger job to get 2d embeddings from embeddings  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.trigger2d_embeddings_job_with_http_info(body, dataset_id, embedding_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param Trigger2dEmbeddingJobRequest body: (required)
        :param MongoObjectID dataset_id: ObjectId of the dataset (required)
        :param MongoObjectID embedding_id: ObjectId of the embedding (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['body', 'dataset_id', 'embedding_id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method trigger2d_embeddings_job" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `trigger2d_embeddings_job`")  # noqa: E501
        # verify the required parameter 'dataset_id' is set
        if self.api_client.client_side_validation and ('dataset_id' not in params or
                                                       params['dataset_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `dataset_id` when calling `trigger2d_embeddings_job`")  # noqa: E501
        # verify the required parameter 'embedding_id' is set
        if self.api_client.client_side_validation and ('embedding_id' not in params or
                                                       params['embedding_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `embedding_id` when calling `trigger2d_embeddings_job`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'dataset_id' in params:
            path_params['datasetId'] = params['dataset_id']  # noqa: E501
        if 'embedding_id' in params:
            path_params['embeddingId'] = params['embedding_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['ApiKeyAuth', 'auth0Bearer']  # noqa: E501

        return self.api_client.call_api(
            '/v1/datasets/{datasetId}/embeddings/{embeddingId}/trigger2dEmbeddingsJob', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type=None,  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)



""" CO2 Regularizer """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import torch
from lightly.loss.memory_bank import MemoryBankModule


class CO2Regularizer(MemoryBankModule):
    """Implementation of the CO2 regularizer [0] for self-supervised learning.

    [0] CO2, 2021, https://arxiv.org/abs/2010.02217

    Attributes:
        alpha:
            Weight of the regularization term.
        t_consistency:
            Temperature used during softmax calculations.
        memory_bank_size:
            Number of negative samples to store in the memory bank.
            Use 0 to use the second batch for negative samples.

    Examples:
        >>> # initialize loss function for MoCo
        >>> loss_fn = NTXentLoss(memory_bank_size=4096)
        >>>
        >>> # initialize CO2 regularizer
        >>> co2 = CO2Regularizer(alpha=1.0, memory_bank_size=4096)
        >>>
        >>> # generate two random trasnforms of images
        >>> t0 = transforms(images)
        >>> t1 = transforms(images)
        >>>
        >>> # feed through the MoCo model
        >>> out0, out1 = model(t0, t1)
        >>> 
        >>> # calculate loss and apply regularizer
        >>> loss = loss_fn(out0, out1) + co2(out0, out1)

    """

    def __init__(self,
                alpha: float = 1,
                t_consistency: float = 0.05,
                memory_bank_size: int = 0):

        super(CO2Regularizer, self).__init__(size=memory_bank_size)
        # try-catch the KLDivLoss construction for backwards compatability
        self.log_target = True
        try:
            self.kl_div = torch.nn.KLDivLoss(
                reduction='batchmean',
                log_target=True
            )
        except TypeError:
            self.log_target = False
            self.kl_div = torch.nn.KLDivLoss(
                reduction='batchmean'
            )

        self.t_consistency = t_consistency
        self.alpha = alpha

    def _get_pseudo_labels(self,
                           out0: torch.Tensor,
                           out1: torch.Tensor,
                           negatives: torch.Tensor = None):
        """Computes the soft pseudo labels across negative samples.

        Args:
            out0:
                Output projections of the first set of transformed images (query).
                Shape: bsz x n_ftrs
            out1:
                Output projections of the second set of transformed images (positive sample).
                Shape: bsz x n_ftrs
            negatives:
                Negative samples to compare against. If this is None, the second
                batch of images will be used as negative samples.
                Shape: memory_bank_size x n_ftrs

        Returns:
            Log probability that a positive samples will classify each negative
            sample as the positive sample.
            Shape: bsz x (bsz - 1) or bsz x memory_bank_size

        """
        batch_size, _ = out0.shape
        if negatives is None:
            # use second batch as negative samples
            # l_pos has shape bsz x 1 and l_neg has shape bsz x bsz
            l_pos = torch.einsum('nc,nc->n', [out0, out1]).unsqueeze(-1)
            l_neg = torch.einsum('nc,ck->nk', [out0, out1.t()])
            # remove elements on the diagonal
            # l_neg has shape bsz x (bsz - 1)
            l_neg = l_neg.masked_select(
                ~torch.eye(batch_size, dtype=bool, device=l_neg.device)
            ).view(batch_size, batch_size - 1)
        else:
            # use memory bank as negative samples
            # l_pos has shape bsz x 1 and l_neg has shape bsz x memory_bank_size
            negatives = negatives.to(out0.device)
            l_pos = torch.einsum('nc,nc->n', [out0, out1]).unsqueeze(-1)
            l_neg = torch.einsum('nc,ck->nk', [out0, negatives.clone().detach()])
            
        # concatenate such that positive samples are at index 0
        logits = torch.cat([l_pos, l_neg], dim=1)
        # divide by temperature
        logits = logits / self.t_consistency

        # the input to kl_div is expected to be log(p) 
        return torch.nn.functional.log_softmax(logits, dim=-1)


    def forward(self,
                out0: torch.Tensor,
                out1: torch.Tensor):
        """Computes the CO2 regularization term for two model outputs.

        Args:
            out0:
                Output projections of the first set of transformed images.
            out1:
                Output projections of the second set of transformed images.

        Returns:
            The regularization term multiplied by the weight factor alpha.

        """

        # normalize the output to length 1
        out0 = torch.nn.functional.normalize(out0, dim=1)
        out1 = torch.nn.functional.normalize(out1, dim=1)

        # ask memory bank for negative samples and extend it with out1 if 
        # out1 requires a gradient, otherwise keep the same vectors in the 
        # memory bank (this allows for keeping the memory bank constant e.g.
        # for evaluating the loss on the test set)
        # if the memory_bank size is 0, negatives will be None
        out1, negatives = \
            super(CO2Regularizer, self).forward(out1, update=True)
        
        # get log probabilities
        p = self._get_pseudo_labels(out0, out1, negatives)
        q = self._get_pseudo_labels(out1, out0, negatives)
        
        # calculate symmetrized kullback leibler divergence
        if self.log_target:
            div = self.kl_div(p, q) + self.kl_div(q, p)
        else:
            # can't use log_target because of early torch version
            div = self.kl_div(p, torch.exp(q)) + self.kl_div(q, torch.exp(p))

        return self.alpha * 0.5 * div



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class SelectionConfig(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'n_samples': 'int',
        'proportion_samples': 'float',
        'strategies': 'list[SelectionConfigEntry]'
    }

    attribute_map = {
        'n_samples': 'nSamples',
        'proportion_samples': 'proportionSamples',
        'strategies': 'strategies'
    }

    def __init__(self, n_samples=None, proportion_samples=None, strategies=None, _configuration=None):  # noqa: E501
        """SelectionConfig - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._n_samples = None
        self._proportion_samples = None
        self._strategies = None
        self.discriminator = None

        if n_samples is not None:
            self.n_samples = n_samples
        if proportion_samples is not None:
            self.proportion_samples = proportion_samples
        self.strategies = strategies

    @property
    def n_samples(self):
        """Gets the n_samples of this SelectionConfig.  # noqa: E501


        :return: The n_samples of this SelectionConfig.  # noqa: E501
        :rtype: int
        """
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n_samples):
        """Sets the n_samples of this SelectionConfig.


        :param n_samples: The n_samples of this SelectionConfig.  # noqa: E501
        :type: int
        """

        self._n_samples = n_samples

    @property
    def proportion_samples(self):
        """Gets the proportion_samples of this SelectionConfig.  # noqa: E501


        :return: The proportion_samples of this SelectionConfig.  # noqa: E501
        :rtype: float
        """
        return self._proportion_samples

    @proportion_samples.setter
    def proportion_samples(self, proportion_samples):
        """Sets the proportion_samples of this SelectionConfig.


        :param proportion_samples: The proportion_samples of this SelectionConfig.  # noqa: E501
        :type: float
        """

        self._proportion_samples = proportion_samples

    @property
    def strategies(self):
        """Gets the strategies of this SelectionConfig.  # noqa: E501


        :return: The strategies of this SelectionConfig.  # noqa: E501
        :rtype: list[SelectionConfigEntry]
        """
        return self._strategies

    @strategies.setter
    def strategies(self, strategies):
        """Sets the strategies of this SelectionConfig.


        :param strategies: The strategies of this SelectionConfig.  # noqa: E501
        :type: list[SelectionConfigEntry]
        """
        if self._configuration.client_side_validation and strategies is None:
            raise ValueError("Invalid value for `strategies`, must not be `None`")  # noqa: E501

        self._strategies = strategies

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(SelectionConfig, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SelectionConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SelectionConfig):
            return True

        return self.to_dict() != other.to_dict()



import io
import csv
import tempfile
import hashlib
from datetime import datetime
from typing import List
from urllib.request import Request, urlopen
from lightly.api.utils import retry

from lightly.openapi_generated.swagger_client import \
    DimensionalityReductionMethod, Trigger2dEmbeddingJobRequest
from lightly.openapi_generated.swagger_client.models.dataset_embedding_data \
    import DatasetEmbeddingData
from lightly.openapi_generated.swagger_client.models.write_csv_url_data \
    import WriteCSVUrlData
from lightly.utils.io import check_filenames, check_embeddings


class EmbeddingDoesNotExistError(ValueError):
    pass


class _UploadEmbeddingsMixin:

    def _get_csv_reader_from_read_url(self, read_url: str):
        """Makes a get request to the signed read url and returns the .csv file.

        """
        request = Request(read_url, method='GET')
        with urlopen(request) as response:
            buffer = io.StringIO(response.read().decode('utf-8'))
            reader = csv.reader(buffer)

        return reader

    def set_embedding_id_to_latest(self):
        """Sets the self.embedding_id to the one of the latest on the server.

        """
        embeddings_on_server: List[DatasetEmbeddingData] = \
            self._embeddings_api.get_embeddings_by_dataset_id(
                dataset_id=self.dataset_id
            )
        if len(embeddings_on_server) == 0:
            raise RuntimeError(f"There are no known embeddings for dataset_id {self.dataset_id}.")
        # return first entry as the API returns newest first
        self.embedding_id = embeddings_on_server[0].id

    def get_embedding_by_name(
            self, name: str, ignore_suffix: bool = True
    ) -> DatasetEmbeddingData:
        """Gets an embedding form the server by name.

        Args:
            name:
                The name of the embedding to get.
            ignore_suffix:
                If true, a suffix of the embedding name on the server
                is ignored.

        Returns:
            The embedding data.

        Raises:
            EmbeddingDoesNotExistError:
                If the name does not match the name of an embedding
                on the server.

        """
        embeddings_on_server: List[DatasetEmbeddingData] = \
            self._embeddings_api.get_embeddings_by_dataset_id(
                dataset_id=self.dataset_id
            )
        try:
            if ignore_suffix:
                embedding = next(
                    embedding for embedding in embeddings_on_server if
                    embedding.name.startswith(name)
                )
            else:
                embedding = next(
                    embedding for embedding in embeddings_on_server if
                    embedding.name == name
                )
        except StopIteration:
            raise EmbeddingDoesNotExistError(
                f"Embedding with the specified name "
                f"does not exist on the server: {name}"
            )
        return embedding

    def upload_embeddings(self, path_to_embeddings_csv: str, name: str):
        """Uploads embeddings to the server.

        First checks that the specified embedding name is not on the server. If it is, the upload is aborted.
        Then creates a new csv with the embeddings in the order specified on the server. Next it uploads it to the server.
        The received embedding_id is saved as a property of self.

        Args:
            path_to_embeddings_csv:
                The path to the .csv containing the embeddings, e.g. "path/to/embeddings.csv"
            name:
                The name of the embedding. If an embedding with such a name already exists on the server,
                the upload is aborted.

        """
        check_embeddings(path_to_embeddings_csv, remove_additional_columns=True)

        # Try to append the embeddings on the server, if they exist
        try:
            embedding = self.get_embedding_by_name(name, ignore_suffix=True)
            # -> append rows from server
            print('Appending embeddings from server.')
            self.append_embeddings(path_to_embeddings_csv, embedding.id)
            now = datetime.now().strftime('%Y%m%d_%Hh%Mm%Ss')
            name = f'{name}_{now}'
        except EmbeddingDoesNotExistError:
            pass

        # create a new csv with the filenames in the desired order
        rows_csv = self._order_csv_by_filenames(
            path_to_embeddings_csv=path_to_embeddings_csv)

        # get the URL to upload the csv to
        response: WriteCSVUrlData = \
            self._embeddings_api.get_embeddings_csv_write_url_by_id(self.dataset_id, name=name)
        self.embedding_id = response.embedding_id
        signed_write_url = response.signed_write_url

        # save the csv rows in a temporary in-memory string file
        # using a csv writer and then read them as bytes
        with tempfile.SpooledTemporaryFile(mode="rw") as f:
            writer = csv.writer(f)
            writer.writerows(rows_csv)
            f.seek(0)
            embeddings_csv_as_bytes = f.read().encode('utf-8')

        # write the bytes to a temporary in-memory byte file
        with tempfile.SpooledTemporaryFile(mode='r+b') as f_bytes:
            f_bytes.write(embeddings_csv_as_bytes)
            f_bytes.seek(0)
            retry(
                self.upload_file_with_signed_url,
                file=f_bytes,
                signed_write_url=signed_write_url
            )

        # trigger the 2d embeddings job
        for dimensionality_reduction_method in [
            DimensionalityReductionMethod.PCA,
            DimensionalityReductionMethod.TSNE,
            DimensionalityReductionMethod.UMAP
        ]:

            body = Trigger2dEmbeddingJobRequest(
                dimensionality_reduction_method=dimensionality_reduction_method)
            self._embeddings_api.trigger2d_embeddings_job(
                body=body,
                dataset_id=self.dataset_id,
                embedding_id=self.embedding_id
            )


    def append_embeddings(self,
                          path_to_embeddings_csv: str,
                          embedding_id: str):
        """Concatenates the embeddings from the server to the local ones.

        Loads the embedding csv file belonging to the embedding_id, and
        appends all of its rows to the local embeddings file located at
        'path_to_embeddings_csv'.

        Args:
            path_to_embeddings_csv:
                The path to the csv containing the local embeddings.
            embedding_id:
                Id of the embedding summary of the embeddings on the server.

        Raises:
            RuntimeError:
                If the number of columns in the local and the remote
                embeddings file mismatch.
        
        """

        # read embedding from API
        embedding_read_url = self._embeddings_api \
            .get_embeddings_csv_read_url_by_id(self.dataset_id, embedding_id)
        embedding_reader = self._get_csv_reader_from_read_url(embedding_read_url)
        rows = list(embedding_reader)
        header, online_rows = rows[0], rows[1:]

        # read local embedding
        with open(path_to_embeddings_csv, 'r') as f:
            local_rows = list(csv.reader(f))

            if len(local_rows[0]) != len(header):
                raise RuntimeError(
                    'Column mismatch! Number of columns in local and remote'
                    f' embeddings files must match but are {len(local_rows[0])}'
                    f' and {len(header)} respectively.'
                )

            local_rows = local_rows[1:]

        # combine online and local embeddings
        total_rows = [header]
        filename_to_local_row = { row[0]: row for row in local_rows }
        for row in online_rows:
            # pick local over online filename if it exists
            total_rows.append(filename_to_local_row.pop(row[0], row))
        # add all local rows which were not added yet
        total_rows.extend(list(filename_to_local_row.values()))
        
        # save embeddings again
        with open(path_to_embeddings_csv, 'w') as f:
            writer = csv.writer(f)
            writer.writerows(total_rows)
        

    def _order_csv_by_filenames(self, path_to_embeddings_csv: str) -> List[str]:
        """Orders the rows in a csv according to the order specified on the server and saves it as a new file.

        Args:
            path_to_embeddings_csv:
                the path to the csv to order

        Returns:
            the filepath to the new csv

        """
        with open(path_to_embeddings_csv, 'r') as f:
            data = csv.reader(f)

            rows = list(data)
            header_row = rows[0]
            rows_without_header = rows[1:]
            index_filenames = header_row.index('filenames')
            filenames = [row[index_filenames] for row in rows_without_header]

            filenames_on_server = self.get_filenames()

            if len(filenames) != len(filenames_on_server):
                raise ValueError(f'There are {len(filenames)} rows in the embedding file, but '
                                 f'{len(filenames_on_server)} filenames/samples on the server.')
            if set(filenames) != set(filenames_on_server):
                raise ValueError(f'The filenames in the embedding file and '
                                 f'the filenames on the server do not align')
            check_filenames(filenames)

            rows_without_header_ordered = \
                self._order_list_by_filenames(filenames, rows_without_header)

            rows_csv = [header_row]
            rows_csv += rows_without_header_ordered

        return rows_csv



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class JobStatusData(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'id': 'MongoObjectID',
        'dataset_id': 'MongoObjectID',
        'status': 'JobState',
        'meta': 'JobStatusMeta',
        'wait_time_till_next_poll': 'int',
        'created_at': 'Timestamp',
        'last_modified_at': 'Timestamp',
        'finished_at': 'Timestamp',
        'error': 'str',
        'result': 'JobStatusDataResult'
    }

    attribute_map = {
        'id': 'id',
        'dataset_id': 'datasetId',
        'status': 'status',
        'meta': 'meta',
        'wait_time_till_next_poll': 'waitTimeTillNextPoll',
        'created_at': 'createdAt',
        'last_modified_at': 'lastModifiedAt',
        'finished_at': 'finishedAt',
        'error': 'error',
        'result': 'result'
    }

    def __init__(self, id=None, dataset_id=None, status=None, meta=None, wait_time_till_next_poll=None, created_at=None, last_modified_at=None, finished_at=None, error=None, result=None, _configuration=None):  # noqa: E501
        """JobStatusData - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._id = None
        self._dataset_id = None
        self._status = None
        self._meta = None
        self._wait_time_till_next_poll = None
        self._created_at = None
        self._last_modified_at = None
        self._finished_at = None
        self._error = None
        self._result = None
        self.discriminator = None

        self.id = id
        if dataset_id is not None:
            self.dataset_id = dataset_id
        self.status = status
        if meta is not None:
            self.meta = meta
        self.wait_time_till_next_poll = wait_time_till_next_poll
        self.created_at = created_at
        if last_modified_at is not None:
            self.last_modified_at = last_modified_at
        if finished_at is not None:
            self.finished_at = finished_at
        if error is not None:
            self.error = error
        if result is not None:
            self.result = result

    @property
    def id(self):
        """Gets the id of this JobStatusData.  # noqa: E501


        :return: The id of this JobStatusData.  # noqa: E501
        :rtype: MongoObjectID
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this JobStatusData.


        :param id: The id of this JobStatusData.  # noqa: E501
        :type: MongoObjectID
        """
        if self._configuration.client_side_validation and id is None:
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def dataset_id(self):
        """Gets the dataset_id of this JobStatusData.  # noqa: E501


        :return: The dataset_id of this JobStatusData.  # noqa: E501
        :rtype: MongoObjectID
        """
        return self._dataset_id

    @dataset_id.setter
    def dataset_id(self, dataset_id):
        """Sets the dataset_id of this JobStatusData.


        :param dataset_id: The dataset_id of this JobStatusData.  # noqa: E501
        :type: MongoObjectID
        """

        self._dataset_id = dataset_id

    @property
    def status(self):
        """Gets the status of this JobStatusData.  # noqa: E501


        :return: The status of this JobStatusData.  # noqa: E501
        :rtype: JobState
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this JobStatusData.


        :param status: The status of this JobStatusData.  # noqa: E501
        :type: JobState
        """
        if self._configuration.client_side_validation and status is None:
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def meta(self):
        """Gets the meta of this JobStatusData.  # noqa: E501


        :return: The meta of this JobStatusData.  # noqa: E501
        :rtype: JobStatusMeta
        """
        return self._meta

    @meta.setter
    def meta(self, meta):
        """Sets the meta of this JobStatusData.


        :param meta: The meta of this JobStatusData.  # noqa: E501
        :type: JobStatusMeta
        """

        self._meta = meta

    @property
    def wait_time_till_next_poll(self):
        """Gets the wait_time_till_next_poll of this JobStatusData.  # noqa: E501

        The time in seconds the client should wait before doing the next poll.  # noqa: E501

        :return: The wait_time_till_next_poll of this JobStatusData.  # noqa: E501
        :rtype: int
        """
        return self._wait_time_till_next_poll

    @wait_time_till_next_poll.setter
    def wait_time_till_next_poll(self, wait_time_till_next_poll):
        """Sets the wait_time_till_next_poll of this JobStatusData.

        The time in seconds the client should wait before doing the next poll.  # noqa: E501

        :param wait_time_till_next_poll: The wait_time_till_next_poll of this JobStatusData.  # noqa: E501
        :type: int
        """
        if self._configuration.client_side_validation and wait_time_till_next_poll is None:
            raise ValueError("Invalid value for `wait_time_till_next_poll`, must not be `None`")  # noqa: E501

        self._wait_time_till_next_poll = wait_time_till_next_poll

    @property
    def created_at(self):
        """Gets the created_at of this JobStatusData.  # noqa: E501


        :return: The created_at of this JobStatusData.  # noqa: E501
        :rtype: Timestamp
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this JobStatusData.


        :param created_at: The created_at of this JobStatusData.  # noqa: E501
        :type: Timestamp
        """
        if self._configuration.client_side_validation and created_at is None:
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def last_modified_at(self):
        """Gets the last_modified_at of this JobStatusData.  # noqa: E501


        :return: The last_modified_at of this JobStatusData.  # noqa: E501
        :rtype: Timestamp
        """
        return self._last_modified_at

    @last_modified_at.setter
    def last_modified_at(self, last_modified_at):
        """Sets the last_modified_at of this JobStatusData.


        :param last_modified_at: The last_modified_at of this JobStatusData.  # noqa: E501
        :type: Timestamp
        """

        self._last_modified_at = last_modified_at

    @property
    def finished_at(self):
        """Gets the finished_at of this JobStatusData.  # noqa: E501


        :return: The finished_at of this JobStatusData.  # noqa: E501
        :rtype: Timestamp
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this JobStatusData.


        :param finished_at: The finished_at of this JobStatusData.  # noqa: E501
        :type: Timestamp
        """

        self._finished_at = finished_at

    @property
    def error(self):
        """Gets the error of this JobStatusData.  # noqa: E501


        :return: The error of this JobStatusData.  # noqa: E501
        :rtype: str
        """
        return self._error

    @error.setter
    def error(self, error):
        """Sets the error of this JobStatusData.


        :param error: The error of this JobStatusData.  # noqa: E501
        :type: str
        """

        self._error = error

    @property
    def result(self):
        """Gets the result of this JobStatusData.  # noqa: E501


        :return: The result of this JobStatusData.  # noqa: E501
        :rtype: JobStatusDataResult
        """
        return self._result

    @result.setter
    def result(self, result):
        """Sets the result of this JobStatusData.


        :param result: The result of this JobStatusData.  # noqa: E501
        :type: JobStatusDataResult
        """

        self._result = result

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(JobStatusData, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, JobStatusData):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, JobStatusData):
            return True

        return self.to_dict() != other.to_dict()



import torch
import torch.distributed as dist
import torch.nn.functional as F

from lightly.utils.dist import gather
from lightly.loss.vicreg_loss import VICRegLoss
from lightly.models.utils import nearest_neighbors
from typing import List, Tuple
import copy


class VICRegLLoss(torch.nn.Module):
    """Implementation of the VICRegL Loss from VICRegL[0] paper.
    This implementation follows the code published by the authors. [1]

    [0] Bardes, A. et. al, 2022, VICReg... https://arxiv.org/abs/2210.01571
    [1] https://github.com/facebookresearch/VICRegL

    Examples:

        >>> # initialize loss function
        >>> loss_fn = VICRegLLoss()
        >>>
        >>> # generate two random transforms of images
        >>> t0 = transforms(images)
        >>> t1 = transforms(images)
        >>>
        >>> # feed through model images
        >>> out0, out1, out_local0, out_local1, grid0, grid1  = model(t0, t1)
        >>>
        >>> # calculate loss
        >>> loss = loss_fn(out0, out1, out_local0, out_local1, grid0, grid1)
    """

    def __init__(
        self,
        lambda_param: float = 25.0,
        mu_param: float = 25.0,
        nu_param: float = 1.0,
        alpha: float = 0.25,
        gather_distributed: bool = False,
        eps: float = 0.0001,
        num_matches: Tuple[int] = (20, 4),
    ):
        """Lambda, mu and nu params configuration with default value like in [0]
        Args:
            lambda_param:
                Coefficient for the invariance term of the loss
                Defaults to 25.0 [0].
            mu_param:
                Coefficient for the variance term of the loss
                Defaults to 25.0 [0].
            nu_param:
                Coefficient for the covariance term of the loss
                Defaults to 1.0 [0].
            alpha:
                Coefficient to weight local loss with global loss
            gather_distributed:
                If True then the cross-correlation matrices from all gpus are
                gathered and summed before the loss calculation.
            eps:
                Numerical epsilon
                Defaults to 0.0001 [1].
            num_matches:
                Number of local matches between patches in the KNN
        """
        super(VICRegLLoss, self).__init__()
        self.lambda_param = lambda_param
        self.mu_param = mu_param
        self.nu_param = nu_param
        self.gather_distributed = gather_distributed
        self.alpha = alpha
        self.eps = eps
        self.num_matches = num_matches
        self.vicregloss = VICRegLoss()

    def _nearest_neighbors_on_l2(
        self, input_maps: torch.Tensor, candidate_maps: torch.Tensor, num_matches: int
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        input_maps: (B, H * W, C)
        candidate_maps: (B, H * W, C)
        """

        distances = torch.cdist(input_maps, candidate_maps)

        return nearest_neighbors(input_maps, candidate_maps, distances, num_matches)

    def _nearest_neighbors_on_grid(
        self,
        input_grid: torch.Tensor,
        candidate_grid: torch.Tensor,
        input_maps: torch.Tensor,
        candidate_maps: torch.Tensor,
        num_matches: int,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        input_grid: (B, H * W, 2)
        candidate_grid: (B, H * W, 2)
        input_maps: (B, H * W, C)
        candidate_maps: (B, H * W, C)
        """

        distances = torch.cdist(input_grid, candidate_grid)

        return nearest_neighbors(input_maps, candidate_maps, distances, num_matches)

    def local_loss(
        self,
        z_global: torch.Tensor,
        z_local: torch.Tensor,
        grid_global: torch.Tensor,
        grid_local: torch.Tensor,
    ) -> torch.Tensor:
        """Compute the local loss

        Compute the local loss between two sets of maps using nearest neighbors and location loss.

        Args:
            z_global:
                A tensor of global maps. It has size: [batch_size, global_image_height_crop, global_image_width_crop, global_feature_dimension]
            z_local:
                A tensor of local maps. It has size: [batch_size, local_image_height_crop, local_image_width_crop, local_feature_dimension]
            grid_global:
                A tensor of grids for the global maps. It has size: [batch_size, grid_size, grid_size, 2]
            grid_local:
                A tensor of grids for the local maps. It has size: [batch_size, grid_size, grid_size, 2]

        Returns:
            A tensor of the local loss between the two sets of maps. It has size: [batch_size]"""

        inv_loss = 0.0

        z_global = z_global.flatten(1, 2)
        z_local = z_local.flatten(1, 2)

        # L2 loss

        z_global_filtered, z_global_nn = self._nearest_neighbors_on_l2(
            input_maps=z_global, candidate_maps=z_local, num_matches=self.num_matches[0]
        )

        z_local_filtered, z_local_nn = self._nearest_neighbors_on_l2(
            input_maps=z_local, candidate_maps=z_global, num_matches=self.num_matches[1]
        )

        # add fast version from paper implementation

        inv_loss_global = F.mse_loss(z_global_filtered, z_global_nn)
        inv_loss_local = F.mse_loss(z_local_filtered, z_local_nn)
        inv_loss = inv_loss + (inv_loss_global / 2 + inv_loss_local / 2)

        grid_global = grid_global.flatten(1, 2)
        grid_local = grid_local.flatten(1, 2)

        # distance loss

        z_global_filtered, z_global_nn = self._nearest_neighbors_on_grid(
            input_grid=grid_global,
            candidate_grid=grid_local,
            input_maps=z_global,
            candidate_maps=z_local,
            num_matches=self.num_matches[0],
        )

        z_local_filtered, z_local_nn = self._nearest_neighbors_on_grid(
            input_grid=grid_local,
            candidate_grid=grid_global,
            input_maps=z_local,
            candidate_maps=z_global,
            num_matches=self.num_matches[1],
        )

        inv_loss_global = F.mse_loss(z_global_filtered, z_global_nn)
        inv_loss_local = F.mse_loss(z_local_filtered, z_local_nn)
        inv_loss = inv_loss + (inv_loss_global / 2 + inv_loss_local / 2)

        local_loss = self.lambda_param * inv_loss

        return local_loss

    def forward(
        self,
        z_global: torch.Tensor,
        z_local: torch.Tensor,
        z_global_local_features: torch.Tensor,
        z_local_local_features: torch.Tensor,
        grid_global: torch.Tensor,
        grid_local: torch.Tensor,
    ) -> torch.Tensor:
        """Compute the overall loss between two sets of maps, using global loss and local loss.

        Computes the global loss using VicRegLoss and local loss using nearest neighbor search.
        It then combines the global and local loss using a scalar value alpha, and returns the result as loss.

        Args:
            z_global:
                A tensor of global maps. It has size: [batch_size, global_image_height, global_image_width, global_feature_dimension]
            z_local:
                A tensor of local maps. It has size: [batch_size, local_image_height, local_image_width, local_feature_dimension]
            z_global_local_features:
                A tensor of local features for the global maps. It has size: [batch_size, global_image_height_crop, global_image_width_crop, global_feature_dimension]
            z_local_local_features:
                A tensor of local features for the local maps. It has size: [batch_size, global_image_height_crop, global_image_width_crop, global_feature_dimension]
            grid_global:
                A tensor of grids for the global maps. It has size: [batch_size, grid_size, grid_size, 2]
            grid_local:
                A tensor of grids for the local maps. It has size: [batch_size, grid_size, grid_size, 2]

        Returns:
            A tensor of the overall loss between the two sets of maps. It has size: [batch_size].
        """

        if z_global_local_features.shape[0] < 1 or z_local_local_features.shape[0] < 1:
            raise ValueError(
                f"z_global_local and z_local_local must have batch size > 1 but found {z_global_local_features.shape[0]} and  {z_local_local_features.shape[0]}"
            )

        global_loss = self.vicregloss.forward(z_a=z_global, z_b=z_local)

        local_loss = self.local_loss(
            z_global=z_global_local_features,
            z_local=z_local_local_features,
            grid_global=grid_global,
            grid_local=grid_local,
        )

        loss = self.alpha * global_loss + (1 - self.alpha) * local_loss

        return loss



""" SimCLR Model """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import warnings

import torch
import torch.nn as nn

from lightly.models.modules import SimCLRProjectionHead


class SimCLR(nn.Module):
    """Implementation of the SimCLR[0] architecture

    Recommended loss: :py:class:`lightly.loss.ntx_ent_loss.NTXentLoss`

    [0] SimCLR, 2020, https://arxiv.org/abs/2002.05709

    Attributes:
        backbone:
            Backbone model to extract features from images.
        num_ftrs:
            Dimension of the embedding (before the projection head).
        out_dim:
            Dimension of the output (after the projection head).

    """

    def __init__(self, backbone: nn.Module, num_ftrs: int = 32, out_dim: int = 128):

        super(SimCLR, self).__init__()

        self.backbone = backbone
        self.projection_head = SimCLRProjectionHead(num_ftrs, num_ftrs, out_dim)

        warnings.warn(
            Warning(
                "The high-level building block SimCLR will be deprecated in version 1.3.0. "
                + "Use low-level building blocks instead. "
                + "See https://docs.lightly.ai/self-supervised-learning/lightly.models.html for more information"
            ),
            PendingDeprecationWarning,
        )

    def forward(
        self, x0: torch.Tensor, x1: torch.Tensor = None, return_features: bool = False
    ):
        """Embeds and projects the input images.

        Extracts features with the backbone and applies the projection
        head to the output space. If both x0 and x1 are not None, both will be
        passed through the backbone and projection head. If x1 is None, only
        x0 will be forwarded.

        Args:
            x0:
                Tensor of shape bsz x channels x W x H.
            x1:
                Tensor of shape bsz x channels x W x H.
            return_features:
                Whether or not to return the intermediate features backbone(x).

        Returns:
            The output projection of x0 and (if x1 is not None) the output
            projection of x1. If return_features is True, the output for each x
            is a tuple (out, f) where f are the features before the projection
            head.

        Examples:
            >>> # single input, single output
            >>> out = model(x)
            >>>
            >>> # single input with return_features=True
            >>> out, f = model(x, return_features=True)
            >>>
            >>> # two inputs, two outputs
            >>> out0, out1 = model(x0, x1)
            >>>
            >>> # two inputs, two outputs with return_features=True
            >>> (out0, f0), (out1, f1) = model(x0, x1, return_features=True)

        """

        # forward pass of first input x0
        f0 = self.backbone(x0).flatten(start_dim=1)
        out0 = self.projection_head(f0)

        # append features if requested
        if return_features:
            out0 = (out0, f0)

        # return out0 if x1 is None
        if x1 is None:
            return out0

        # forward pass of second input x1
        f1 = self.backbone(x1).flatten(start_dim=1)
        out1 = self.projection_head(f1)

        # append features if requested
        if return_features:
            out1 = (out1, f1)

        # return both outputs
        return out0, out1



""" Jigsaw Crop """

# Copyright (c) 2021. Lightly AG and its affiliates.
# All Rights Reserved

import torch
from torchvision import transforms
import numpy as np
from PIL import Image


class Jigsaw(object):
    """Implementation of Jigsaw image augmentation, inspired from PyContrast library.
    
    Generates n_grid**2 random crops and returns a list.
    
    This augmentation is instrumental to PIRL.
    
    Attributes:
        n_grid:
            Side length of the meshgrid, sqrt of the number of crops.
        img_size:
            Size of image.
        crop_size:
            Size of crops.
        transform:
            Transformation to apply on each crop.
    
    Examples:
        >>> from lightly.transforms import Jigsaw
        >>>
        >>> jigsaw_crop = Jigsaw(n_grid=3, img_size=255, crop_size=64, transform=transforms.ToTensor())
        >>>
        >>> # img is a PIL image
        >>> crops = jigsaw_crops(img)
    """
    def __init__(self, n_grid=3, img_size=255, crop_size=64, transform=transforms.ToTensor()):
        self.n_grid = n_grid
        self.img_size = img_size
        self.crop_size = crop_size
        self.grid_size = int(img_size / self.n_grid)
        self.side = self.grid_size - self.crop_size
        self.transform = transform

        yy, xx = np.meshgrid(np.arange(n_grid), np.arange(n_grid))
        self.yy = np.reshape(yy * self.grid_size, (n_grid * n_grid,))
        self.xx = np.reshape(xx * self.grid_size, (n_grid * n_grid,))

    def __call__(self, img):
        """Performs the Jigsaw augmentation
        Args:
            img:
                PIL image to perform Jigsaw augmentation on.

        Returns:
            Torch tensor with stacked crops.
        """
        r_x = np.random.randint(0, self.side + 1, self.n_grid * self.n_grid)
        r_y = np.random.randint(0, self.side + 1, self.n_grid * self.n_grid)
        img = np.asarray(img, np.uint8)
        crops = []
        for i in range(self.n_grid * self.n_grid):
            crops.append(img[self.xx[i] + r_x[i]: self.xx[i] + r_x[i] + self.crop_size,
                         self.yy[i] + r_y[i]: self.yy[i] + r_y[i] + self.crop_size, :])
        crops = [Image.fromarray(crop) for crop in crops]
        crops = torch.stack([self.transform(crop) for crop in crops])
        crops = crops[np.random.permutation(self.n_grid ** 2)]
        return crops



from abc import ABC
from typing import Optional, List

from lightly.openapi_generated.swagger_client import TaskType


class PredictionSingletonRepr(ABC):
    def __init__(
        self,
        type: str,
        taskName: str,
        categoryId: int,
        score: float,
        cropDatasetId: Optional[str] = None,
        cropSampleId: Optional[str] = None,
    ):
        self.type = type
        self.taskName = taskName
        self.categoryId = categoryId
        self.score = score
        self.cropDatasetId = cropDatasetId
        self.cropSampleId = cropSampleId

    def to_dict(self):
        return {key: value for key, value in vars(self).items() if value is not None}


class PredictionSingletonClassificationRepr(PredictionSingletonRepr):
    def __init__(
        self,
        taskName: str,
        categoryId: int,
        score: float,
        probabilities: Optional[List[float]] = None,
    ):
        super().__init__(
            type=TaskType.CLASSIFICATION,
            taskName=taskName,
            categoryId=categoryId,
            score=score,
        )
        self.probabilities = probabilities


class PredictionSingletonObjectDetectionRepr(PredictionSingletonRepr):
    def __init__(
        self,
        taskName: str,
        categoryId: int,
        score: float,
        bbox: List[int],
        cropDatasetId: Optional[str] = None,
        cropSampleId: Optional[str] = None,
        probabilities: Optional[List[float]] = None,
    ):
        super().__init__(
            type=TaskType.OBJECT_DETECTION,
            taskName=taskName,
            categoryId=categoryId,
            score=score,
            cropDatasetId=cropDatasetId,
            cropSampleId=cropSampleId,
        )
        self.bbox = bbox
        self.probabilities = probabilities


class PredictionSingletonSemanticSegmentationRepr(PredictionSingletonRepr):
    def __init__(
        self,
        taskName: str,
        categoryId: int,
        score: float,
        segmentation: List[int],
        probabilities: Optional[List[float]] = None,
    ):
        super().__init__(
            type=TaskType.SEMANTIC_SEGMENTATION,
            taskName=taskName,
            categoryId=categoryId,
            score=score,
        )
        self.segmentation = segmentation
        self.probabilities = probabilities


# Not used
class PredictionSingletonInstanceSegmentationRepr(PredictionSingletonRepr):
    def __init__(
        self,
        taskName: str,
        categoryId: int,
        score: float,
        cropDatasetId: str,
        cropSampleId: str,
        segmentation: str,
        probabilities: Optional[List[float]] = None,
    ):
        super().__init__(
            type=TaskType.INSTANCE_SEGMENTATION,
            taskName=taskName,
            categoryId=categoryId,
            score=score,
            cropDatasetId=cropDatasetId,
            cropSampleId=cropSampleId,
        )
        self.segmentation = segmentation
        self.probabilities = probabilities


# Not used
class PredictionSingletonKeypointDetectionRepr(PredictionSingletonRepr):
    def __init__(
        self,
        taskName: str,
        categoryId: int,
        score: float,
        keypoints: List[int],
        probabilities: Optional[List[float]] = None,
    ):
        super().__init__(
            type=TaskType.KEYPOINT_DETECTION,
            taskName=taskName,
            categoryId=categoryId,
            score=score,
        )
        self.keypoints = keypoints
        self.probabilities = probabilities



""" Barlow Twins resnet-based Model [0]
[0] Zbontar,J. et.al. 2021. Barlow Twins... https://arxiv.org/abs/2103.03230
"""

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

import warnings

import torch
import torch.nn as nn

from lightly.models.modules import BarlowTwinsProjectionHead


class BarlowTwins(nn.Module):
    """Implementation of BarlowTwins[0] network.

    Recommended loss: :py:class:`lightly.loss.barlow_twins_loss.BarlowTwinsLoss`

    Default params are the ones explained in the original paper [0].
    [0] Zbontar,J. et.al. 2021. Barlow Twins... https://arxiv.org/abs/2103.03230

    Attributes:
        backbone:
            Backbone model to extract features from images.
            ResNet-50 in original paper [0].
        num_ftrs:
            Dimension of the embedding (before the projection head).
        proj_hidden_dim:
            Dimension of the hidden layer of the projection head. This should
            be the same size as `num_ftrs`.
        out_dim:
            Dimension of the output (after the projection head).

    """

    def __init__(
        self,
        backbone: nn.Module,
        num_ftrs: int = 2048,
        proj_hidden_dim: int = 8192,
        out_dim: int = 8192,
    ):

        super(BarlowTwins, self).__init__()

        self.backbone = backbone
        self.num_ftrs = num_ftrs
        self.proj_hidden_dim = proj_hidden_dim
        self.out_dim = out_dim

        self.projection_mlp = BarlowTwinsProjectionHead(
            num_ftrs, proj_hidden_dim, out_dim
        )

        warnings.warn(
            Warning(
                "The high-level building block BarlowTwins will be deprecated in version 1.3.0. "
                + "Use low-level building blocks instead. "
                + "See https://docs.lightly.ai/self-supervised-learning/lightly.models.html for more information"
            ),
            PendingDeprecationWarning,
        )

    def forward(
        self, x0: torch.Tensor, x1: torch.Tensor = None, return_features: bool = False
    ):

        """Forward pass through BarlowTwins.

        Extracts features with the backbone and applies the projection
        head to the output space. If both x0 and x1 are not None, both will be
        passed through the backbone and projection. If x1 is None, only x0 will
        be forwarded.
        Barlow Twins only implement a projection head unlike SimSiam.

        Args:
            x0:
                Tensor of shape bsz x channels x W x H.
            x1:
                Tensor of shape bsz x channels x W x H.
            return_features:
                Whether or not to return the intermediate features backbone(x).

        Returns:
            The output projection of x0 and (if x1 is not None)
            the output projection of x1. If return_features is
            True, the output for each x is a tuple (out, f) where f are the
            features before the projection head.

        Examples:
            >>> # single input, single output
            >>> out = model(x)
            >>>
            >>> # single input with return_features=True
            >>> out, f = model(x, return_features=True)
            >>>
            >>> # two inputs, two outputs
            >>> out0, out1 = model(x0, x1)
            >>>
            >>> # two inputs, two outputs with return_features=True
            >>> (out0, f0), (out1, f1) = model(x0, x1, return_features=True)
        """
        # forward pass first input
        f0 = self.backbone(x0).flatten(start_dim=1)
        out0 = self.projection_mlp(f0)

        # append features if requested
        if return_features:
            out0 = (out0, f0)

        if x1 is None:
            return out0

        # forward pass second input
        f1 = self.backbone(x1).flatten(start_dim=1)
        out1 = self.projection_mlp(f1)

        # append features if requested
        if return_features:
            out1 = (out1, f1)

        return out0, out1



from typing import List, Union
from PIL import Image

import torch
import torchvision

from lightly.data.collate import BaseCollateFunction, MultiViewCollateFunction

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = ModuleNotFoundError(
        "Matplotlib is not installed on your system. Please install it to use the plotting"
        "functionalities. See https://matplotlib.org/ for installation instructions."
    )

def _check_matplotlib_available() -> None:
    if isinstance(plt, Exception):
        raise plt


@torch.no_grad()
def std_of_l2_normalized(z: torch.Tensor) -> torch.Tensor:
    """Calculates the mean of the standard deviation of z along each dimension.

    This measure was used by [0] to determine the level of collapse of the
    learned representations. If the returned number is 0., the outputs z have
    collapsed to a constant vector. "If the output z has a zero-mean isotropic
    Gaussian distribution" [0], the returned number should be close to 1/sqrt(d)
    where d is the dimensionality of the output.

    [0]: https://arxiv.org/abs/2011.10566

    Args:
        z:
            A torch tensor of shape batch_size x dimension.

    Returns:
        The mean of the standard deviation of the l2 normalized tensor z along
        each dimension.
    
    """

    if len(z.shape) != 2:
        raise ValueError(
            f'Input tensor must have two dimensions but has {len(z.shape)}!'
        )

    z_norm = torch.nn.functional.normalize(z, dim=1)
    return torch.std(z_norm, dim=0).mean()


def apply_transform_without_normalize(
    image: Image.Image,
    transform,
):
    """Applies the transform to the image but skips ToTensor and Normalize.

    """
    skippable_transforms = (
        torchvision.transforms.ToTensor,
        torchvision.transforms.Normalize,
    )
    if isinstance(transform, torchvision.transforms.Compose):
        for transform_ in transform.transforms:
            image = apply_transform_without_normalize(image, transform_)
    elif not isinstance(transform, skippable_transforms):
        image = transform(image)
    return image


def generate_grid_of_augmented_images(
    input_images: List[Image.Image],
    collate_function: Union[BaseCollateFunction, MultiViewCollateFunction],
):
    """Returns a grid of augmented images. Images in a column belong together.

    This function ignores the transforms ToTensor and Normalize for visualization purposes.

    Args:
        input_images:
            List of PIL images for which the augmentations should be plotted.
        collate_function:
            The collate function of the self-supervised learning algorithm.
            Must be of type BaseCollateFunction or MultiViewCollateFunction.

    Returns:
        A grid of augmented images. Images in a column belong together.

    """
    grid = []
    if isinstance(collate_function, BaseCollateFunction):
        for _ in range(2):
            grid.append([
                apply_transform_without_normalize(image, collate_function.transform)
                for image in input_images
            ])
    elif isinstance(collate_function, MultiViewCollateFunction):
        for transform in collate_function.transforms:
            grid.append([
                apply_transform_without_normalize(image, transform)
                for image in input_images
            ])
    else:
        raise ValueError(
            'Collate function must be one of '
            '(BaseCollateFunction, MultiViewCollateFunction) '
            f'but is {type(collate_function)}.'
        )
    return grid


def plot_augmented_images(
    input_images: List[Image.Image],
    collate_function: Union[BaseCollateFunction, MultiViewCollateFunction],
):
    """Returns a figure showing original images in the left column and augmented images to their right.

    This function ignores the transforms ToTensor and Normalize for visualization purposes.

    Args:
        input_images:
            List of PIL images for which the augmentations should be plotted.
        collate_function:
            The collate function of the self-supervised learning algorithm.
            Must be of type BaseCollateFunction or MultiViewCollateFunction.

    Returns:
        A figure showing the original images in the left column and the augmented
        images to their right. If the collate_function is an instance of the
        BaseCollateFunction, two example augmentations are shown. For
        MultiViewCollateFunctions all the generated views are shown.

    """

    _check_matplotlib_available()

    if len(input_images) == 0:
        raise ValueError('There must be at least one input image.')

    grid = generate_grid_of_augmented_images(input_images, collate_function)
    grid.insert(0, input_images)
    nrows = len(input_images)
    ncols = len(grid)

    fig, axs = plt.subplots(nrows, ncols, figsize=(ncols * 1.5, nrows * 1.5))

    for i in range(nrows):
        for j in range(ncols):
            ax = axs[i][j] if len(input_images) > 1 else axs[j]
            img = grid[j][i]
            ax.imshow(img)
            ax.set_axis_off()

    ax_top_left = axs[0, 0] if len(input_images) > 1 else axs[0]
    ax_top_left.set(title='Original images')
    ax_top_left.title.set_size(8)
    ax_top_next = axs[0, 1] if len(input_images) > 1 else axs[1]
    ax_top_next.set(title='Augmented images')
    ax_top_next.title.set_size(8)
    fig.tight_layout()

    return fig



""" BaseEmbeddings """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved
import copy
import os

import omegaconf
from omegaconf import DictConfig
from pytorch_lightning import LightningModule, Trainer

from lightly.embedding import callbacks


class BaseEmbedding(LightningModule):
    """All trainable embeddings must inherit from BaseEmbedding.

    """

    def __init__(self,
                 model,
                 criterion,
                 optimizer,
                 dataloader,
                 scheduler=None):
        """ Constructor

        Args:
            model: (torch.nn.Module)
            criterion: (torch.nn.Module)
            optimizer: (torch.optim.Optimizer)
            dataloader: (torch.utils.data.DataLoader)

        """

        super(BaseEmbedding, self).__init__()
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.dataloader = dataloader
        self.scheduler = scheduler
        self.checkpoint = None
        self.cwd = os.getcwd()

    def forward(self, x0, x1):
        return self.model(x0, x1)

    def training_step(self, batch, batch_idx):

        # get the two image transformations
        (x0, x1), _, _ = batch
        # forward pass of the transformations
        y0, y1 = self(x0, x1)
        # calculate loss
        loss = self.criterion(y0, y1)
        # log loss and return
        self.log('loss', loss)
        return loss

    def configure_optimizers(self):
        if self.scheduler is None:
            return self.optimizer
        else:
            return [self.optimizer], [self.scheduler]

    def train_dataloader(self):
        return self.dataloader

    def train_embedding(
        self,
        trainer_config: DictConfig,
        checkpoint_callback_config: DictConfig,
        summary_callback_config: DictConfig,
    ):
        """ Train the model on the provided dataset.

        Args:
            trainer_config: pylightning_trainer arguments, examples include:
                min_epochs: (int) Minimum number of epochs to train
                max_epochs: (int) Maximum number of epochs to train
                gpus: (int) Number of gpus to use
                enable_model_summary: (bool) Whether to enable model summarisation.
                weights_summary: (str) DEPRECATED. How to print a summary of the model and weights.
            checkpoint_callback_config: ModelCheckpoint callback arguments
            summary_callback_config: ModelSummary callback arguments

        Returns:
            A trained encoder, ready for embedding datasets.

        """
        trainer_callbacks = []

        checkpoint_cb = callbacks.create_checkpoint_callback(**checkpoint_callback_config)
        trainer_callbacks.append(checkpoint_cb)

        summary_cb = callbacks.create_summary_callback(
            summary_callback_config=summary_callback_config,
            trainer_config=trainer_config,
        )
        if summary_cb is not None:
            trainer_callbacks.append(summary_cb)

        # Remove weights_summary from trainer_config now that the summary callback
        # has been created. TODO: Drop support for the "weights_summary" argument.
        trainer_config_copy = copy.deepcopy(trainer_config)
        if "weights_summary" in trainer_config_copy:
            with omegaconf.open_dict(trainer_config_copy):
                del trainer_config_copy["weights_summary"]

        trainer = Trainer(**trainer_config_copy, callbacks=trainer_callbacks)

        trainer.fit(self)

        if checkpoint_cb.best_model_path != "":
            self.checkpoint = os.path.join(self.cwd, checkpoint_cb.best_model_path)

    def embed(self, *args, **kwargs):
        """Must be implemented by classes which inherit from BaseEmbedding.

        """
        raise NotImplementedError()



""" Lightly Model Zoo """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

ZOO = {

    'resnet-9/simclr/d16/w0.0625':
    'https://storage.googleapis.com/models_boris/whattolabel-resnet9-simclr-d16-w0.0625-i-ce0d6bd9.pth',

    'resnet-9/simclr/d16/w0.125':
    'https://storage.googleapis.com/models_boris/whattolabel-resnet9-simclr-d16-w0.125-i-7269c38d.pth',

    'resnet-18/simclr/d16/w1.0':
    'https://storage.googleapis.com/models_boris/whattolabel-resnet18-simclr-d16-w1.0-i-58852cb9.pth',

    'resnet-18/simclr/d32/w1.0':
    'https://storage.googleapis.com/models_boris/whattolabel-resnet18-simclr-d32-w1.0-i-085d0693.pth',

    'resnet-34/simclr/d16/w1.0':
    'https://storage.googleapis.com/models_boris/whattolabel-resnet34-simclr-d16-w1.0-i-6e80d963.pth',

    'resnet-34/simclr/d32/w1.0':
    'https://storage.googleapis.com/models_boris/whattolabel-resnet34-simclr-d32-w1.0-i-9f185b45.pth'

}

def checkpoints():
    """Returns the Lightly model zoo as a list of checkpoints.

    Checkpoints:
        ResNet-9:
            SimCLR with width = 0.0625 and num_ftrs = 16
        ResNet-9:
            SimCLR with width = 0.125 and num_ftrs = 16
        ResNet-18:
            SimCLR with width = 1.0 and num_ftrs = 16
        ResNet-18:
            SimCLR with width = 1.0 and num_ftrs = 32
        ResNet-34:
            SimCLR with width = 1.0 and num_ftrs = 16
        ResNet-34:
            SimCLR with width = 1.0 and num_ftrs = 32

    Returns:
        A list of available checkpoints as URLs.

    """
    return [item for key, item in ZOO.items()]




from __future__ import annotations
from functools import partial
import math
from typing import Callable, List, Optional

import torch
import torch.nn as nn
from lightly.models import utils

# vision_transformer requires torchvision >= 0.12
from torchvision.models import vision_transformer
from torchvision.models.vision_transformer import ConvStemConfig


class MAEEncoder(vision_transformer.Encoder):
    """Encoder for the Masked Autoencoder model [0].

    Encodes patch embeddings. Code inspired by [1].

    - [0]: Masked Autoencoder, 2021, https://arxiv.org/abs/2111.06377
    - [1]: https://github.com/facebookresearch/mae

    Attributes:
        seq_length:
            Token sequence length, including the class token.
        num_layers:
            Number of transformer blocks.
        num_heads:
            Number of attention heads.
        hidden_dim:
            Dimension of the input and output tokens.
        mlp_dim:
            Dimension of the MLP in the transformer block.
        dropout:
            Percentage of elements set to zero after the MLP in the transformer.
        attention_dropout:
            Percentage of elements set to zero after the attention head.

    """
    def __init__(
        self, 
        seq_length: int, 
        num_layers: int, 
        num_heads: int, 
        hidden_dim: int,
        mlp_dim: int, 
        dropout: float, 
        attention_dropout: float, 
        norm_layer: Callable[..., torch.nn.Module] = partial(nn.LayerNorm, eps=1e-6),
    ):
        super().__init__(
            seq_length=seq_length,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            norm_layer=norm_layer,
        )

    @classmethod
    def from_vit_encoder(cls, vit_encoder: vision_transformer.Encoder) -> MAEEncoder:
        """Creates a MAEEncoder from a torchvision ViT encoder."""
        # Create a new instance with dummy values as they will be overwritten
        # by the copied vit_encoder attributes
        encoder = cls(
            seq_length=1,
            num_layers=1,
            num_heads=1,
            hidden_dim=1,
            mlp_dim=1,
            dropout=0,
            attention_dropout=0,
        )
        encoder.pos_embedding = vit_encoder.pos_embedding
        encoder.dropout = vit_encoder.dropout
        encoder.layers = vit_encoder.layers
        encoder.ln = vit_encoder.ln
        return encoder

    def forward(
        self, 
        input: torch.Tensor, 
        idx_keep: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Encode input tokens.

        Args:
            input:
                Batch of token sequences.
            idx_keep:
                Tensor with shape (batch_size, num_tokens_to_keep) where each
                entry is an index of the token to keep in the respective batch.
                If specified, only the indexed tokens will be encoded.

        Returns:
            Batch of encoded output tokens.
        """
        input = input + self.interpolate_pos_encoding(input)
        if idx_keep is not None:
            input = utils.get_at_index(input, idx_keep)
        return self.ln(self.layers(self.dropout(input)))

    def interpolate_pos_encoding(self, input: torch.Tensor):
        """Returns the interpolated positional embedding for the given input.

        This function interpolates self.pos_embedding for all tokens in the input,
        ignoring the class token. This allows encoding variable sized images.

        Args:
            input: 
               Input tensor with shape (batch_size, num_sequences).

        """
        # code copied from:
        # https://github.com/facebookresearch/msn/blob/4388dc1eadbe3042b85d3296d41b9b207656e043/src/deit.py#L291
        npatch = input.shape[1] - 1
        N = self.pos_embedding.shape[1] - 1
        if npatch == N:
            return self.pos_embedding
        class_emb = self.pos_embedding[:, 0]
        pos_embedding = self.pos_embedding[:, 1:]
        dim = input.shape[-1]
        pos_embedding = nn.functional.interpolate(
            pos_embedding.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
            scale_factor=math.sqrt(npatch / N),
            mode='bicubic',
        )
        pos_embedding = pos_embedding.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_emb.unsqueeze(0), pos_embedding), dim=1)


class MAEBackbone(vision_transformer.VisionTransformer):
    """Backbone for the Masked Autoencoder model [0].

    Converts images into patches and encodes them. Code inspired by [1]. 
    Note that this implementation uses a learned positional embedding while [0]
    uses a fixed positional embedding.

    - [0]: Masked Autoencoder, 2021, https://arxiv.org/abs/2111.06377
    - [1]: https://github.com/facebookresearch/mae
    - [2]: Early Convolutions Help Transformers See Better, 2021, https://arxiv.org/abs/2106.14881.

    Attributes:
        image_size:
            Input image size.
        patch_size:
            Width and height of the image patches. image_size must be a multiple
            of patch_size.
        num_layers:
            Number of transformer blocks.
        num_heads:
            Number of attention heads.
        hidden_dim:
            Dimension of the input and output tokens.
        mlp_dim:
            Dimension of the MLP in the transformer block.
        dropout:
            Percentage of elements set to zero after the MLP in the transformer.
        attention_dropout:
            Percentage of elements set to zero after the attention head.
        num_classes:
            Number of classes for the classification head. Currently not used.
        representation_size:
            If specified, an additional linear layer is added before the 
            classification head to change the token dimension from hidden_dim
            to representation_size. Currently not used.
        norm_layer:
            Callable that creates a normalization layer.
        conv_stem_configs:
            If specified, a convolutional stem is added at the beggining of the
            network following [2]. Not used in the original Masked Autoencoder
            paper [0].

    """
    def __init__(
        self, 
        image_size: int, 
        patch_size: int, 
        num_layers: int, 
        num_heads: int, 
        hidden_dim: int, 
        mlp_dim: int, 
        dropout: float = 0, 
        attention_dropout: float = 0, 
        num_classes: int = 1000, 
        representation_size: Optional[int] = None, 
        norm_layer: Callable[..., torch.nn.Module] = partial(nn.LayerNorm, eps=1e-6),
        conv_stem_configs: Optional[List[ConvStemConfig]] = None
    ):
        super().__init__(
            image_size=image_size,
            patch_size=patch_size,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            num_classes=num_classes,
            representation_size=representation_size,
            norm_layer=norm_layer,
            conv_stem_configs=conv_stem_configs,
        )
        self.encoder = MAEEncoder(
            seq_length=self.seq_length,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            norm_layer=norm_layer,
        )

    @classmethod
    def from_vit(cls, vit: vision_transformer.VisionTransformer) -> MAEBackbone:
        """Creates a MAEBackbone from a torchvision ViT model."""
        # Create a new instance with dummy values as they will be overwritten
        # by the copied vit_encoder attributes
        backbone = cls(
            image_size=vit.image_size,
            patch_size=vit.patch_size,
            num_layers=1,
            num_heads=1,
            hidden_dim=vit.hidden_dim,
            mlp_dim=vit.mlp_dim,
            dropout=vit.dropout,
            attention_dropout=vit.attention_dropout,
            num_classes=vit.num_classes,
            representation_size=vit.representation_size,
            norm_layer=vit.norm_layer,
        )
        backbone.conv_proj = vit.conv_proj
        backbone.class_token = vit.class_token
        backbone.seq_length = vit.seq_length
        backbone.heads = vit.heads
        backbone.encoder = MAEEncoder.from_vit_encoder(vit.encoder)
        return backbone

    def forward(
        self, 
        images: torch.Tensor, 
        idx_keep: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Returns encoded class tokens from a batch of images.

        Args:
            images:
                Tensor with shape (batch_size, channels, image_size, image_size).
            idx_keep:
                Tensor with shape (batch_size, num_tokens_to_keep) where each
                entry is an index of the token to keep in the respective batch.
                If specified, only the indexed tokens will be passed to the
                encoder.
        
        Returns:
            Tensor with shape (batch_size, hidden_dim) containing the 
            encoded class token for every image.

        """
        out = self.encode(images, idx_keep)
        class_token = out[:, 0]
        return class_token

    def encode(
        self, 
        images: torch.Tensor, 
        idx_keep: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Returns encoded class and patch tokens from images.

        Args:
            images:
                Tensor with shape (batch_size, channels, image_size, image_size).
            idx_keep:
                Tensor with shape (batch_size, num_tokens_to_keep) where each
                entry is an index of the token to keep in the respective batch.
                If specified, only the indexed tokens will be passed to the
                encoder.
        
        Returns:
            Tensor with shape (batch_size, sequence_length, hidden_dim) 
            containing the encoded class and patch tokens for every image.

        """
        out = self.images_to_tokens(images, prepend_class_token=True)
        return self.encoder(out, idx_keep)


    def images_to_tokens(self, images: torch.Tensor, prepend_class_token: bool) -> torch.Tensor:
        """Converts images into patch tokens.
        
        Args:
            images:
                Tensor with shape (batch_size, channels, image_size, image_size).
        
        Returns:
            Tensor with shape (batch_size, sequence_length - 1, hidden_dim)
            containing the patch tokens.
        """
        x = self.conv_proj(images)
        tokens =  x.flatten(2).transpose(1, 2)
        if prepend_class_token:
            tokens = utils.prepend_class_token(tokens, self.class_token)
        return tokens


class MAEDecoder(vision_transformer.Encoder):
    """Decoder for the Masked Autoencoder model [0].

    Decodes encoded patches and predicts pixel values for every patch.
    Code inspired by [1].

    - [0]: Masked Autoencoder, 2021, https://arxiv.org/abs/2111.06377
    - [1]: https://github.com/facebookresearch/mae

    Attributes:
        seq_length:
            Token sequence length, including the class token.
        num_layers:
            Number of transformer blocks.
        num_heads:
            Number of attention heads.
        embed_input_dim:
            Dimension of the input tokens. Usually be equal to the hidden
            dimension of the MAEEncoder or MAEBackbone.
        hidden_dim:
            Dimension of the decoder tokens.
        mlp_dim:
            Dimension of the MLP in the transformer block.
        out_dim:
            Output dimension of the prediction for a single patch. Usually equal
            to (3 * patch_size ** 2).
        dropout:
            Percentage of elements set to zero after the MLP in the transformer.
        attention_dropout:
            Percentage of elements set to zero after the attention head.

    """
    def __init__(
        self,
        seq_length: int,
        num_layers: int,
        num_heads: int,
        embed_input_dim: int,
        hidden_dim: int,
        mlp_dim: int,
        out_dim: int,
        dropout: float = 0.0,
        attention_dropout: float = 0.0,
        norm_layer: Callable[..., nn.Module] = partial(nn.LayerNorm, eps=1e-6),
    ):
        super().__init__(
            seq_length=seq_length,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            norm_layer=norm_layer,
        )
        self.decoder_embed = nn.Linear(embed_input_dim, hidden_dim, bias=True)
        self.prediction_head = nn.Linear(hidden_dim, out_dim)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Returns predicted pixel values from encoded tokens.

        Args:
            input:
                Tensor with shape (batch_size, seq_length, embed_input_dim).

        Returns:
            Tensor with shape (batch_size, seq_length, out_dim).

        """
        out = self.embed(input)
        out = self.decode(out)
        return self.predict(out)

    def embed(self, input: torch.Tensor) -> torch.Tensor:
        """Embeds encoded input tokens into decoder token dimension.

        This is a single linear layer that changes the token dimension from
        embed_input_dim to hidden_dim.

        Args:
            input:
                Tensor with shape (batch_size, seq_length, embed_input_dim)
                containing the encoded tokens.

        Returns:
            Tensor with shape (batch_size, seq_length, hidden_dim) containing
            the embedded tokens.

        """
        return self.decoder_embed(input)

    def decode(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass through the decoder transformer.

        Args:
            input:
                Tensor with shape (batch_size, seq_length, hidden_dim) containing
                the encoded tokens.

        Returns:
            Tensor with shape (batch_size, seq_length, hidden_dim) containing
            the decoded tokens.

        """
        return super().forward(input)

    def predict(self, input: torch.Tensor) -> torch.Tensor:
        """Predics pixel values from decoded tokens.

        Args:
            input:
                Tensor with shape (batch_size, seq_length, hidden_dim) containing
                the decoded tokens.

        Returns:
            Tensor with shape (batch_size, seq_length, out_dim) containing
            predictions for each token.

        """
        return self.prediction_head(input)



# coding: utf-8

"""
    Lightly API

    Lightly.ai enables you to do self-supervised learning in an easy and intuitive way. The lightly.ai OpenAPI spec defines how one can interact with our REST API to unleash the full potential of lightly.ai  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: support@lightly.ai
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from lightly.openapi_generated.swagger_client.configuration import Configuration


class DockerWorkerConfigV2LightlyLoader(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'batch_size': 'int',
        'shuffle': 'bool',
        'num_workers': 'int',
        'drop_last': 'bool'
    }

    attribute_map = {
        'batch_size': 'batchSize',
        'shuffle': 'shuffle',
        'num_workers': 'numWorkers',
        'drop_last': 'dropLast'
    }

    def __init__(self, batch_size=None, shuffle=None, num_workers=None, drop_last=None, _configuration=None):  # noqa: E501
        """DockerWorkerConfigV2LightlyLoader - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._batch_size = None
        self._shuffle = None
        self._num_workers = None
        self._drop_last = None
        self.discriminator = None

        if batch_size is not None:
            self.batch_size = batch_size
        if shuffle is not None:
            self.shuffle = shuffle
        if num_workers is not None:
            self.num_workers = num_workers
        if drop_last is not None:
            self.drop_last = drop_last

    @property
    def batch_size(self):
        """Gets the batch_size of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501


        :return: The batch_size of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :rtype: int
        """
        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size):
        """Sets the batch_size of this DockerWorkerConfigV2LightlyLoader.


        :param batch_size: The batch_size of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :type: int
        """

        self._batch_size = batch_size

    @property
    def shuffle(self):
        """Gets the shuffle of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501


        :return: The shuffle of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :rtype: bool
        """
        return self._shuffle

    @shuffle.setter
    def shuffle(self, shuffle):
        """Sets the shuffle of this DockerWorkerConfigV2LightlyLoader.


        :param shuffle: The shuffle of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :type: bool
        """

        self._shuffle = shuffle

    @property
    def num_workers(self):
        """Gets the num_workers of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501


        :return: The num_workers of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :rtype: int
        """
        return self._num_workers

    @num_workers.setter
    def num_workers(self, num_workers):
        """Sets the num_workers of this DockerWorkerConfigV2LightlyLoader.


        :param num_workers: The num_workers of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :type: int
        """

        self._num_workers = num_workers

    @property
    def drop_last(self):
        """Gets the drop_last of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501


        :return: The drop_last of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :rtype: bool
        """
        return self._drop_last

    @drop_last.setter
    def drop_last(self, drop_last):
        """Sets the drop_last of this DockerWorkerConfigV2LightlyLoader.


        :param drop_last: The drop_last of this DockerWorkerConfigV2LightlyLoader.  # noqa: E501
        :type: bool
        """

        self._drop_last = drop_last

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(DockerWorkerConfigV2LightlyLoader, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DockerWorkerConfigV2LightlyLoader):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DockerWorkerConfigV2LightlyLoader):
            return True

        return self.to_dict() != other.to_dict()



""" Utility method for comparing versions of libraries """

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved


def version_compare(v0: str, v1: str):
    """Returns 1 if version of v0 is larger than v1 and -1 otherwise
    
    Use this method to compare Python package versions and see which one is
    newer.

    Examples:

        >>> # compare two versions
        >>> version_compare('1.2.0', '1.1.2')
        >>> 1
    """
    v0 = [int(n) for n in v0.split('.')][::-1]
    v1 = [int(n) for n in v1.split('.')][::-1]
    if len(v0) != 3 or len(v1) != 3:
        raise ValueError(
            f'Length of version strings is not 3 (expected pattern `x.y.z`) but is '
            f'{v0} and {v1}.'
        )
    pairs = list(zip(v0, v1))[::-1]
    for x, y in pairs:
        if x < y:
            return -1
        if x > y:
            return 1
    return 0



""" The lightly.cli module provides a console interface
    for training self-supervised models, embedding,
    and filtering datasets
"""

# Copyright (c) 2020. Lightly AG and its affiliates.
# All Rights Reserved

from lightly.cli.lightly_cli import lightly_cli
from lightly.cli.train_cli import train_cli
from lightly.cli.embed_cli import embed_cli
from lightly.cli.upload_cli import upload_cli
from lightly.cli.download_cli import download_cli
from lightly.cli.crop_cli import crop_cli
